/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   Protocol services - ISO layer
   Copyright (C) Matthew Chapman 1999-2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rdesktop.h"

RD_BOOL g_extended_data_supported = False;
extern int g_use_rdp;

/* Send a self-contained ISO PDU */
static void
iso_send_msg(uint8 code)
{
	STREAM s;

	s = tcp_init(11);

	out_uint8(s, 3);	/* version */
	out_uint8(s, 0);	/* reserved */
	out_uint16_be(s, 11);	/* length */

	out_uint8(s, 6);	/* hdrlen */
	out_uint8(s, code);
	out_uint16(s, 0);	/* dst_ref */
	out_uint16(s, 0);	/* src_ref */
	out_uint8(s, 0);	/* class */

	s_mark_end(s);
	tcp_send(s);
}

/* Send a disconnect provider ultimatum PDU */
static void
iso_send_disconnect()
{
	STREAM s;

	s = tcp_init(9);

	out_uint8(s, 0x03);	/* version */
	out_uint8(s, 0x00);	/* reserved */
	out_uint16_be(s, 0x09);	/* length */

	out_uint8(s, 0x02);	/* hdrlen */
	out_uint8(s, 0xf0);
	out_uint8(s, 0x80);
	out_uint8(s, 0x21);
	out_uint8(s, 0x80);

	s_mark_end(s);
	tcp_send(s);
}

/* Send X.224 connection request PDU */
static void
iso_send_connection_request(char *username, RD_BOOL send_token)
{
	STREAM s;
	int length;
	
	if (send_token)
		length = 30 + strlen(username) + (g_use_rdp != RDP4 ? 8 : 0);
	else
		length = 11 + (g_use_rdp != RDP4 ? 8 : 0);

	s = tcp_init(length);

	out_uint8(s, 3);	/* version */
	out_uint8(s, 0);	/* reserved */
	out_uint16_be(s, length);	/* length */

	out_uint8(s, length - 5);	/* hdrlen */
	out_uint8(s, ISO_PDU_CR);
	out_uint16(s, 0);	/* dst_ref */
	out_uint16(s, 0);	/* src_ref */
	out_uint8(s, 0);	/* class */

	if (send_token)
	{
		out_uint8p(s, "Cookie: mstshash=", strlen("Cookie: mstshash="));
		out_uint8p(s, username, strlen(username));
		out_uint8(s, 0x0d);	/* carrage-return */
		out_uint8(s, 0x0a);	/* line-feed */
	}

	/* RDP negotiation request */
	if (g_use_rdp != RDP4)
	{
		out_uint8(s, TYPE_RDP_NEG_REQ);	/* type */
		out_uint8(s, 0x00);	/* flags */
		out_uint16_le(s, 0x0008);	/* length */
		out_uint32_le(s, PROTOCOL_RDP_FLAG);	/* requested protocol */
	}

	s_mark_end(s);
	tcp_send(s);
}

/* Receive a message on the ISO layer, return code */
static STREAM
iso_recv_msg(uint8 * code, uint8 * rdpver)
{
	STREAM s;
	uint16 length;
	uint8 version;

	s = tcp_recv(NULL, 4);
	if (s == NULL)
		return NULL;
	in_uint8(s, version);
	if (rdpver != NULL)
		*rdpver = version;
	if (version == 3)
	{
		in_uint8s(s, 1);	/* pad */
		in_uint16_be(s, length);
	}
	else
	{
		in_uint8(s, length);
		if (length & 0x80)
		{
			length &= ~0x80;
			next_be(s, length);
		}
	}
	if (length < 4)
	{
		error("Bad packet header\n");
		return NULL;
	}
	s = tcp_recv(s, length - 4);
	if (s == NULL)
		return NULL;
	if (version != 3)
		return s;
	in_uint8s(s, 1);	/* hdrlen */
	in_uint8(s, *code);
	if (*code == ISO_PDU_DT)
	{
		in_uint8s(s, 1);	/* eot */
		return s;
	}
	in_uint8s(s, 5);	/* dst_ref, src_ref, class */
	return s;
}

/* Initialise ISO transport data packet */
STREAM
iso_init(int length)
{
	STREAM s;

	s = tcp_init(length + 7);
	s_push_layer(s, iso_hdr, 7);

	return s;
}

/* Send an ISO data PDU */
void
iso_send(STREAM s)
{
	uint16 length;

	s_pop_layer(s, iso_hdr);
	length = s->end - s->p;

	out_uint8(s, 3);	/* version */
	out_uint8(s, 0);	/* reserved */
	out_uint16_be(s, length);

	out_uint8(s, 2);	/* hdrlen */
	out_uint8(s, ISO_PDU_DT);	/* code */
	out_uint8(s, 0x80);	/* eot */

	tcp_send(s);
}

/* Receive ISO transport data packet */
STREAM
iso_recv(uint8 * rdpver)
{
	STREAM s;
	uint8 code = 0;

	s = iso_recv_msg(&code, rdpver);
	if (s == NULL)
		return NULL;
	if (rdpver != NULL)
		if (*rdpver != 3)
			return s;
	if (code != ISO_PDU_DT)
	{
		error("expected DT, got 0x%x\n", code);
		return NULL;
	}
	return s;
}

/* Check for an optional RDP negotiation response field */
static RD_BOOL
iso_connect_confirm(STREAM s)
{
	int len = s->end - s->data;
	uint8 type;
	uint8 flags;
	uint16 length;
	uint32 protocol;

	if (g_use_rdp == RDP4)
		return True;
	
	/* Check if negotiation response has been received */
	if (len == 19)
	{
		in_uint8(s, type);
		in_uint8(s, flags);
		in_uint16_le(s, length);
		in_uint32_le(s, protocol);

		switch (type)
		{
			case TYPE_RDP_NEG_RSP:	/* 0x02 */
				if ((flags & EXTENDED_CLIENT_DATA_SUPPORTED) ==
					EXTENDED_CLIENT_DATA_SUPPORTED)
					g_extended_data_supported = True;
				switch (protocol)
				{
					case PROTOCOL_RDP_FLAG:
						return True;
					default:
						error("Connection Confirm PDU, Protocol still not "
							"supported: 0x%x\n", protocol);
						return False;
				}
				break;
			case TYPE_RDP_NEG_FAILURE:	/* 0x03 */
				error("Connection Confirm PDU, error 0x%x received\n",
					protocol);
				return False;
			default:
				unimpl("Connection Confirm PDU, type 0x%x\n", type);
				return False;
		}
	}
	else if (len == 11)
		return True;
	else {
		error("Unexpected Connection Confirm PDU length: %d\n", len);
		return False;
	}
}

/* Establish a connection up to the ISO layer */
RD_BOOL
iso_connect(char *server, char *username, RD_BOOL send_token)
{
	STREAM s;
	uint8 code = 0;

	if (!tcp_connect(server))
		return False;

	iso_send_connection_request(username, send_token);

	s = iso_recv_msg(&code, NULL);
	if (s == NULL)
		return False;

	if (code != ISO_PDU_CC)
	{
		error("expected CC, got 0x%x\n", code);
		tcp_disconnect();
		return False;
	}

	return iso_connect_confirm(s);
}

/* Establish a reconnection up to the ISO layer */
RD_BOOL
iso_reconnect(char *server)
{
	STREAM s;
	uint8 code = 0;

	if (!tcp_connect(server))
		return False;

	iso_send_msg(ISO_PDU_CR);

	s = iso_recv_msg(&code, NULL);
	if (s == NULL)
		return False;

	if (code != ISO_PDU_CC)
	{
		error("expected CC, got 0x%x\n", code);
		tcp_disconnect();
		return False;
	}

	return iso_connect_confirm(s);
}

/* Disconnect from the ISO layer */
void
iso_disconnect(int shutdown)
{
	if (shutdown == True)
		iso_send_disconnect();
	else
		iso_send_msg(ISO_PDU_DR);
	tcp_disconnect();
}

/* reset the state to support reconnecting */
void
iso_reset_state(void)
{
	tcp_reset_state();
}
