/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   RDP licensing negotiation
   Copyright (C) Matthew Chapman 1999-2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rdesktop.h"
#include "ssl.h"

extern char g_username[64];
extern char g_hostname[16];
extern int g_use_rdp;
extern uint32 g_server_public_key_len;
extern uint8 g_exponent[SEC_EXPONENT_SIZE];
extern uint8 g_modulus[SEC_MAX_MODULUS_SIZE];

static uint8 g_licence_key[16];
static uint8 g_licence_sign_key[16];
static SERVER_VERSION server_version= { 0, 0};

RD_BOOL g_licence_issued = False;

/* Generate a session key and RC4 keys, given client and server randoms */
static void
licence_generate_keys(uint8 * client_random, uint8 * server_random, uint8 * pre_master_secret)
{
	uint8 master_secret[48];
	uint8 key_block[48];

	/* Generate master secret and then key material */
	sec_hash_48(master_secret, pre_master_secret, client_random, server_random, 'A');
	sec_hash_48(key_block, master_secret, server_random, client_random, 'A');

	/* Store first 16 bytes of session key as MAC secret */
	memcpy(g_licence_sign_key, key_block, 16);

	/* Generate RC4 key from next 16 bytes */
	sec_hash_16(g_licence_key, &key_block[16], client_random, server_random);
}

static void
licence_generate_hwid(uint8 * hwid)
{
	uint8 md5[MD5_DIGEST_LENGTH];
	int i;

	hwid[ 0]= 0;
	hwid[ 1]= 0;
	hwid[ 2]= 0x01;
	switch (g_use_rdp)
	{
		case RDP5:
			hwid[ 3]= 0x03;
			break;
		case RDP6:
			hwid[ 3]= 0x04;
			break;
		default:
			hwid[ 3]= 0x02;
	}

	/* Create a better hardware idetification based on the hostname */
	MD5((unsigned char*)g_hostname, strlen(g_hostname), md5);
	for(i= 0; i < MD5_DIGEST_LENGTH; i++)
		hwid[i+ 4]= md5[i];
}

/* Present an existing licence to the server */
static void
licence_present(uint8 * client_random, uint8 * rsa_data,
		uint8 * licence_data, int licence_size, uint8 * hwid, uint8 * signature)
{
	uint32 sec_flags = SEC_LICENCE_NEG;
	uint16 length =
		16 + SEC_RANDOM_SIZE + SEC_MODULUS_SIZE + SEC_PADDING_SIZE +
		licence_size + LICENCE_HWID_SIZE + LICENCE_SIGNATURE_SIZE;
	uint8 encrypted_premaster_secret[SEC_MODULUS_SIZE+ SEC_PADDING_SIZE];
	STREAM s;

	s = sec_init(sec_flags, length + 4);

	out_uint8(s, LICENCE_TAG_PRESENT);
	if (g_use_rdp == RDP4) {
		out_uint8(s, 0x02);	/* version 2 (4.0) */
	}
	else {
		out_uint8(s, 0x83);	/* version 3 (5.0, 5.2, 6.0) with extended error */
	}
	out_uint16_le(s, length);

	out_uint32_le(s, 0x1); /* KEY_EXCHANGE_ALG_RSA */
	out_uint16(s, 0);

	switch (g_use_rdp)
	{
		case RDP5:
			out_uint16_le(s, 0x0301);
			break;
		case RDP6:
			out_uint16_le(s, 0x0401);
			break;
		default:
			out_uint16_le(s, 0x0201);
	}

	out_uint8p(s, client_random, SEC_RANDOM_SIZE);

	/* encrypted premaster secret */
	memset(encrypted_premaster_secret, 0, SEC_MODULUS_SIZE + SEC_PADDING_SIZE);
	ssl_rsa_encrypt(encrypted_premaster_secret, rsa_data, SEC_PREMASTER_SIZE,
	                g_server_public_key_len, g_modulus, g_exponent);
	out_uint16_le(s, 0x0002);
	out_uint16_le(s, (SEC_MODULUS_SIZE + SEC_PADDING_SIZE));
	out_uint8p(s, encrypted_premaster_secret,
	           SEC_MODULUS_SIZE+ SEC_PADDING_SIZE);
	/* out_uint8s(s, SEC_PADDING_SIZE); */

	out_uint16_le(s, 0x0001);
	out_uint16_le(s, licence_size);
	out_uint8p(s, licence_data, licence_size);

	out_uint16_le(s, 0x0001);
	out_uint16_le(s, LICENCE_HWID_SIZE);
	out_uint8p(s, hwid, LICENCE_HWID_SIZE);

	out_uint8p(s, signature, LICENCE_SIGNATURE_SIZE);

	s_mark_end(s);
	sec_send(s, sec_flags);
}

/* Send a licence request packet */
static void
licence_send_request(uint8 * client_random, uint8 * rsa_data, char *user, char *host)
{
	uint32 sec_flags = SEC_LICENCE_NEG;
	uint16 userlen = strlen(user) + 1;
	uint16 hostlen = strlen(host) + 1;
	uint16 length = 128 + userlen + hostlen;
	STREAM s;

#ifdef IGELDEBUG
	fprintf(stdout, "entered: license_send_request\n");
#endif /* IGELDEBUG */

	s = sec_init(sec_flags, length + 2);

	out_uint8(s, LICENCE_TAG_REQUEST);
	if (g_use_rdp == RDP4) {
		out_uint8(s, 0x02);	/* version 2 (4.0) */
	}
	else {
		out_uint8(s, 0x83);	/* version 3 (5.0, 5.2, 6.0) with extended error */
	}
	out_uint16_le(s, length);

	out_uint32_le(s, 0x1); /* KEY_EXCHANGE_ALG_RSA */
	out_uint16(s, 0);

	switch (g_use_rdp)
	{
		case RDP5:
			out_uint16_le(s, 0x0301);
			break;
		case RDP6:
			out_uint16_le(s, 0x0401);
			break;
		default:
			out_uint16_le(s, 0x0201);
	}

	out_uint8p(s, client_random, SEC_RANDOM_SIZE);
	out_uint16_le(s, 0x0002); /* BB_RANDOM_BLOB */
	out_uint16_le(s, (SEC_MODULUS_SIZE + SEC_PADDING_SIZE));
	out_uint8p(s, rsa_data, SEC_MODULUS_SIZE + SEC_PADDING_SIZE);

	out_uint16_le(s, LICENCE_TAG_USER);
	out_uint16_le(s, userlen);
	out_uint8p(s, user, userlen);

	out_uint16_le(s, LICENCE_TAG_HOST);
	out_uint16_le(s, hostlen);
	out_uint8p(s, host, hostlen);

	s_mark_end(s);
	sec_send(s, sec_flags);
}

/* Read out product informations */
static void
licence_read_product_info(STREAM s)
{
	uint32 length;

	in_uint16_le(s, server_version.minor);
	in_uint16_le(s, server_version.major);
	in_uint32_le(s, length);
	in_uint8s(s, length);
	in_uint32_le(s, length);
	in_uint8s(s, length);
}

/* Process a licence demand packet */
static void
licence_process_demand(STREAM s, uint16 length)
{
	uint8 client_random[SEC_MODULUS_SIZE];
	uint8 premaster_secret[SEC_PREMASTER_SIZE];
	uint8 *server_random;
	uint8 signature[LICENCE_SIGNATURE_SIZE];
	uint8 hwid[LICENCE_HWID_SIZE];
	uint8 rsa_data[64+ SEC_PADDING_SIZE]; /* keylen (e.g. 512bit) / 8 + 8 */
	uint8 *licence_data;
	uint16 length16;
	uint16 type;
	uint32 value;
	int licence_size;
	SSL_RC4 crypt_key;

	/* Retrieve the server random from the incoming packet */
	in_uint8p(s, server_random, SEC_RANDOM_SIZE);

	/* create client random */
	/* if (RAND_bytes(client_random, SEC_RANDOM_SIZE) != 1) */
		memset(client_random, 0, SEC_RANDOM_SIZE);

	/* create premaster secret */
	/* if (RAND_bytes(premaster_secret, SEC_PREMASTER_SIZE) != 1) */
		memset(premaster_secret, 0, SEC_PREMASTER_SIZE);

	memset(rsa_data, 0, sizeof(rsa_data));

	/* check for certificate and public key */
	licence_read_product_info(s);
	in_uint16_le(s, type);
	in_uint16_le(s, length16);
	if ((type != 0x000D) || (length16 != 0x0004))
	{
		warning("Unexpected key exchange structure\n");
		in_uint32_le(s, value);
	}
	else
	{
		in_uint32_le(s, value); /* should be 0x00000001 - KEY_EXCHANGE_ALG_RSA */
	}
	in_uint16_le(s, type);
	in_uint16_le(s, length16);
	if ((type == 0x0003) || (type == 0xf548))
	{
		/* read license and get public key - later ;-) */
		in_uint8s(s, length16);
	}
	else if (length != 0)
	{
		warning("Unexpected license structure\n");
		in_uint8s(s, length16);
	}

#ifdef IGELDEBUG
	fprintf(stdout, "server random:\n");
	hexdump(server_random, SEC_RANDOM_SIZE);
	fprintf(stdout, "client random:\n");
	hexdump(client_random, SEC_RANDOM_SIZE);
	fprintf(stdout, "premaster-secret:\n");
	hexdump(premaster_secret, SEC_PREMASTER_SIZE);
	fprintf(stdout, "exponent:\n");
	hexdump(g_exponent, SEC_EXPONENT_SIZE);
	fprintf(stdout, "modulus:\n");
	hexdump(g_modulus, g_server_public_key_len);
#endif /* IGELDEBUG */

	licence_generate_keys(client_random, server_random, premaster_secret);

	/* Workaround for windows server 2008 and per device licensing */
	if (!((server_version.major == 6) && (server_version.minor == 0)))
	{
		licence_size = load_licence(&licence_data);
		if (licence_size > 0)
		{
			/* Generate a signature for the HWID buffer */
			licence_generate_hwid(hwid);
			sec_sign(signature, 16, g_licence_sign_key, 16, hwid, sizeof(hwid));

			/* Now encrypt the HWID */
			ssl_rc4_set_key(&crypt_key, g_licence_key, 16);
			ssl_rc4_crypt(&crypt_key, hwid, hwid, sizeof(hwid));

			licence_present(client_random, rsa_data, licence_data,
			                licence_size, hwid, signature);
			xfree(licence_data);
			return;
		}
	}

	licence_send_request(client_random, rsa_data, g_username, g_hostname);
}

/* Send an authentication response packet */
static void
licence_send_authresp(uint8 * token, uint8 * crypt_hwid, uint8 * signature)
{
	uint32 sec_flags = SEC_LICENCE_NEG;
	uint16 length = 58;
	STREAM s;

	s = sec_init(sec_flags, length + 2);

	out_uint8(s, LICENCE_TAG_AUTHRESP);
	if (g_use_rdp == RDP4) {
		out_uint8(s, 0x02);	/* version 2 (4.0) */
	}
	else {
		out_uint8(s, 0x83);	/* version 3 (5.0, 5.2, 6.0) with extended error */
	}
	out_uint16_le(s, length);

	out_uint16_le(s, 1);
	out_uint16_le(s, LICENCE_TOKEN_SIZE);
	out_uint8p(s, token, LICENCE_TOKEN_SIZE);

	out_uint16_le(s, 1);
	out_uint16_le(s, LICENCE_HWID_SIZE);
	out_uint8p(s, crypt_hwid, LICENCE_HWID_SIZE);

	out_uint8p(s, signature, LICENCE_SIGNATURE_SIZE);

	s_mark_end(s);
	sec_send(s, sec_flags);
}

/* Parse an authentication request packet */
static RD_BOOL
licence_parse_authreq(STREAM s, uint8 ** token, uint8 ** signature)
{
	uint16 tokenlen;

	in_uint8s(s, 6);	/* unknown: f8 3d 15 00 04 f6 */

	in_uint16_le(s, tokenlen);
	if (tokenlen != LICENCE_TOKEN_SIZE)
	{
		error("token len %d\n", tokenlen);
		return False;
	}

	in_uint8p(s, *token, tokenlen);
	in_uint8p(s, *signature, LICENCE_SIGNATURE_SIZE);

	return s_check_end(s);
}

/* Process an authentication request packet */
static void
licence_process_authreq(STREAM s)
{
	uint8 *in_token = NULL, *in_sig;
	uint8 out_token[LICENCE_TOKEN_SIZE], decrypt_token[LICENCE_TOKEN_SIZE];
	uint8 hwid[LICENCE_HWID_SIZE], crypt_hwid[LICENCE_HWID_SIZE];
	uint8 sealed_buffer[LICENCE_TOKEN_SIZE + LICENCE_HWID_SIZE];
	uint8 out_sig[LICENCE_SIGNATURE_SIZE];
	SSL_RC4 crypt_key;

	/* Parse incoming packet and save the encrypted token */
	licence_parse_authreq(s, &in_token, &in_sig);
	memcpy(out_token, in_token, LICENCE_TOKEN_SIZE);

	/* Decrypt the token. It should read TEST in Unicode. */
	ssl_rc4_set_key(&crypt_key, g_licence_key, 16);
	ssl_rc4_crypt(&crypt_key, in_token, decrypt_token, LICENCE_TOKEN_SIZE);

	/* Generate a signature for a buffer of token and HWID */
	licence_generate_hwid(hwid);
	memcpy(sealed_buffer, decrypt_token, LICENCE_TOKEN_SIZE);
	memcpy(sealed_buffer + LICENCE_TOKEN_SIZE, hwid, LICENCE_HWID_SIZE);
	sec_sign(out_sig, 16, g_licence_sign_key, 16, sealed_buffer, sizeof(sealed_buffer));

	/* Now encrypt the HWID */
	ssl_rc4_set_key(&crypt_key, g_licence_key, 16);
	ssl_rc4_crypt(&crypt_key, hwid, crypt_hwid, LICENCE_HWID_SIZE);

	licence_send_authresp(out_token, crypt_hwid, out_sig);
}

/* Process an licence issue packet */
static void
licence_process_issue(STREAM s)
{
	SSL_RC4 crypt_key;
	uint32 length;
	uint32 version;
	unsigned char* plain_license;
	int	plain_license_length;
	int i;

	in_uint8s(s, 2);	/* should be set to 0x0009 since server 2008 */
	in_uint16_le(s, length);
	if (!s_check_rem(s, length))
		return;

	ssl_rc4_set_key(&crypt_key, g_licence_key, 16);
	ssl_rc4_crypt(&crypt_key, s->p, s->p, length);
#ifdef IGELDEBUG
	DEBUG(("Decrypted license:\n"));
	hexdump(s->p, length);
#endif /* IGELDEBUG */

	if ((plain_license = xmalloc( sizeof( unsigned char) * length)) == NULL)
		return;
	memcpy( plain_license, s->p, length);
	plain_license_length= length;

	/* 
	 * Minor mumber must NOT be 0, because, for example, 5.2 is a valid license
	 * version. So the license version would be read and may be later checked,
	 * but is not longer an exit cause.
	 */
	in_uint32_le(s, version);

	g_licence_issued = True;

	/* check for valid license */
	for (i = 0; i < 4; i++)
	{
		in_uint32_le(s, length);
		if (!s_check_rem(s, length))
			return;
		in_uint8s(s, length);
	}

	g_licence_issued = True;
	save_licence(plain_license, plain_license_length);
	xfree( plain_license);
}

/* handle alert/error messages */
static void
licence_error(STREAM s) {
	uint32 error_code;
	uint32 state_transition;

	in_uint32_le(s, error_code);
	in_uint32_le(s, state_transition);

	/* print error messages */
	switch (error_code)
	{
		case LICENCE_ERR_INVALID_SERVER_CERTIFICATE: /* 0x00000001 */
			error("The server certificate is invalid\n");
			break;

		case LICENCE_ERR_NO_LICENSE: /* 0x00000002 */
			error("There is no license\n");
			break;

		case LICENCE_ERR_INVALID_MAC: /* 0x00000003 */
			error("Invalic message authentication code\n");
			break;

		case LICENCE_ERR_INVALID_SCOPE: /* 0x00000004 */
			error("Invalid scope\n");
			break;

		case LICENCE_ERR_NO_LICENSE_SERVER: /* 0x00000006 */
			error("No license server available\n");
			break;

		case LICENCE_STATUS_VALID_CLIENT: /* 0x00000007 */
			if (state_transition == LICENCE_ST_NO_TRANSITION)
				return;
			break;

		case LICENCE_ERR_INVALID_CLIENT: /* 0x00000008 */
			error("Invalid client\n");
			break;

		case LICENCE_ERR_INVALID_PRODUCTID: /* 0x0000000B */
			error("Invalid product id\n");
			break;

		case LICENCE_ERR_INVALID_MESSAGE_LEN: /* 0x0000000C */
			error("Invalid message length\n");
			break;

		default:
			unimpl("licence error code 0x%08x\n", error_code);
	}

	/* state to transition */
	switch (state_transition)
	{
		case LICENCE_ST_TOTAL_ABORT: /* 0x00000001 */
			/* workaround for windows server 2008 */
			if (error_code == LICENCE_ERR_INVALID_CLIENT)
				return;
			else
			{
				error("Program terminates\n");
				exit(1);
			}
			break;

		case LICENCE_ST_NO_TRANSITION: /* 0x00000002 */
			/* do nothing */
			break;

		case LICENCE_ST_PHASE_TO_START: /* 0x00000003 */
			/* should be implemented later */
			break;

		case LICENCE_ST_RESEND_LAST_MESSAGE: /* 0x00000004 */
			/* should be implemented later */
			break;

		default:
			unimpl("licence state to transition code 0x%08x\n", state_transition);
	}
}

/* Process a licence packet */
void
licence_process(STREAM s)
{
	uint8 tag;
	uint8 version;
	uint16 length;

	in_uint8(s, tag);
	in_uint8(s, version);
	in_uint16_le(s, length);

#ifdef IGELDEBUG
	fprintf( stdout, "Received license message: %u, version: %u, length: %u\n",
	         tag, version, length);
	hexdump(s->p- 4, length);
#endif /* IGELDEBUG */

	switch (tag)
	{
		case LICENCE_TAG_DEMAND: /* 0x01 - license request */
			licence_process_demand(s, length);
			break;

		case LICENCE_TAG_AUTHREQ: /* 0x02 - platform challenge */
			licence_process_authreq(s);
			break;

		case LICENCE_TAG_ISSUE: /* 0x03 - new license */
			licence_process_issue(s);
			break;

		case LICENCE_TAG_REISSUE: /* 0x04 - upgrade license */
			licence_process_issue(s);
			break;

		case LICENCE_TAG_RESULT: /* 0xff - error or alert */
			licence_error(s);
			break;

		default:
			unimpl("licence tag 0x%x\n", tag);
	}
}
